# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utils for geometry library."""

from typing import List

import jax.numpy as jnp


def unstack(value: jnp.ndarray, axis: int = -1) -> List[jnp.ndarray]:
  return [jnp.squeeze(v, axis=axis)
          for v in jnp.split(value, value.shape[axis], axis=axis)]
